/*
 * DAVCommunity.java
 *
 * Version: $Revision: 1.5 $
 *
 * Date: $Date: 2005/04/20 14:22:36 $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.app.dav;

import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.ListIterator;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import org.jdom.*;

import org.dspace.authorize.AuthorizeException;
import org.dspace.content.Bitstream;
import org.dspace.content.BitstreamFormat;
import org.dspace.content.Collection;
import org.dspace.content.Community;
import org.dspace.content.DSpaceObject;
import org.dspace.core.Constants;
import org.dspace.core.Context;
import org.dspace.core.Utils;

/**
 * This defines the behavior of DSpace "resources" in the WebDAV
 * interface; it maps DAV operations onto DSpace object.s
 */
class DAVCommunity extends DAVDSpaceObject
{
    /** log4j category */
    private static Logger log = Logger.getLogger(DAVCommunity.class);

    private Community community = null;

    private static final Element short_descriptionProperty =
        new Element("short_description", DAV.NS_DSPACE);
    private static final Element introductory_textProperty =
        new Element("introductory_text", DAV.NS_DSPACE);
    private static final Element side_bar_textProperty =
        new Element("side_bar_text", DAV.NS_DSPACE);
    private static final Element copyright_textProperty =
        new Element("copyright_text", DAV.NS_DSPACE);
    private static final Element logoProperty =
        new Element("logo", DAV.NS_DSPACE);

    private static List allProps = new Vector(commonProps);
    static
    {
        allProps.add(logoProperty);
        allProps.add(short_descriptionProperty);
        allProps.add(introductory_textProperty);
        allProps.add(side_bar_textProperty);
        allProps.add(copyright_textProperty);
        allProps.add(handleProperty);
    }

    protected DAVCommunity(Context context,
                   HttpServletRequest request,
                   HttpServletResponse response,
                   String pathElt[],
                   Community community)
    {
        super(context, request, response, pathElt, community);
        this.community = community;
        this.type = TYPE_COMMUNITY;
    }

    protected List getAllProperties()
    {
        return allProps;
    }

    protected DAVResource[] children()
        throws SQLException
    {
        Vector result = new Vector();
        Community subs[] = community.getSubcommunities();
        for (int i = 0; i < subs.length; i++)
            result.add(new DAVCommunity(context, request, response,
                                        makeChildPath(subs[i]), subs[i]));
        Collection colls[] = community.getCollections();
        for (int i = 0; i < colls.length; i++)
            result.add(new DAVCollection(context, request, response,
                                         makeChildPath(colls[i]), colls[i]));
        return (DAVResource[]) result.toArray(new DAVResource[result.size()]);
    }

    protected Element propfindInternal(Element property)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        String value = null;

        // displayname - title or handle.
        if (elementsEqualIsh(property, displaynameProperty))
        {
            value = getObjectMetadata("name");
            if (value == null)
                value = community.getHandle();
        }

        // special case, value is XML, not string:
        else if (elementsEqualIsh(property, logoProperty))
        {
            Bitstream lbs = community.getLogo();
            Element le;
            if (lbs != null &&
                (le = DAVBitstream.makeXmlBitstream(lbs, this)) != null)
            {
                Element p = new Element("logo", DAV.NS_DSPACE);
                p.addContent(le);
                return p;
            }
        }

        else if (elementsEqualIsh(property, handleProperty))
            value = canonicalizeHandle(community.getHandle());

        else if (elementsEqualIsh(property, short_descriptionProperty))
            value = getObjectMetadata("short_description");

        else if (elementsEqualIsh(property, introductory_textProperty))
            value = getObjectMetadata("introductory_text");

        else if (elementsEqualIsh(property, side_bar_textProperty))
            value = getObjectMetadata("side_bar_text");

        else if (elementsEqualIsh(property, copyright_textProperty))
            value = getObjectMetadata("copyright_text");

        else
            return super.propfindInternal(property);

        // value was set up by "if" clause:
        if (value == null)
            throw new DAVStatusException(HttpServletResponse.SC_NOT_FOUND, "Not found.");
        Element p = new Element(property.getName(), property.getNamespace());
        p.setText(filterForXML(value));
        return p;
    }

    // syntactic sugar around getting community metadata values:
    private String getObjectMetadata(String mdname)
    {
        try
        {
            return community.getMetadata(mdname);
        }
        catch (IllegalArgumentException e)
        {
            return null;
        }
    }

    protected int proppatchInternal(int action, Element prop)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        String newValue = (action == DAV.PROPPATCH_REMOVE) ? null :
                            prop.getText();

        // these are "metadata" values..
        if (elementsEqualIsh(prop, short_descriptionProperty) ||
            elementsEqualIsh(prop, introductory_textProperty) ||
            elementsEqualIsh(prop, side_bar_textProperty) ||
            elementsEqualIsh(prop, copyright_textProperty))
            community.setMetadata(prop.getName(), newValue);

        // displayname maps to "name" metadata:
        else if (elementsEqualIsh(prop, displaynameProperty))
            community.setMetadata("name", newValue);

        // value should be a dspace:bitstream object with inline value.
        else if (elementsEqualIsh(prop, logoProperty))
        {
            if (action == DAV.PROPPATCH_REMOVE)
                community.setLogo(null);
            else
            {
                Element bs = prop.getChild("bitstream", DAV.NS_DSPACE);
                if (bs != null)
                {
                    InputStream bis = DAVBitstream.getXmlBitstreamContent(context, bs);
                    BitstreamFormat bsf = DAVBitstream.getXmlBitstreamFormat(context, bs);
                    if (bis == null || bsf == null)
                        throw new DAVStatusException(DAV.SC_CONFLICT,
                            "Unacceptable value for logo property.");
                    Bitstream nbs = community.setLogo(bis);
                    nbs.setFormat(bsf);
                    nbs.update();
                }
                else
                    throw new DAVStatusException(DAV.SC_CONFLICT,
                        "No <bitstream> element value found for logo property.");
            }
        }

        // anything else
        else
            throw new DAVStatusException(DAV.SC_CONFLICT, "The "+
                prop.getName()+" property cannot be changed.");

        community.update();
        return HttpServletResponse.SC_OK;
    }

    protected void get()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED,
                    "GET method not implemented for Community.");
    }

    protected void put()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED,
                    "PUT method not implemented for Community.");
    }

    protected int copyInternal(DAVResource destination, int depth,
                                      boolean overwrite,
                                      boolean keepProperties)
        throws DAVStatusException, SQLException, AuthorizeException, IOException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED,
            "COPY method not implemented for Collection.");
    }
}
