/*
 * DAVItem.java
 *
 * Version: $Revision: 1.5 $
 *
 * Date: $Date: 2005/04/20 14:22:36 $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.app.dav;

import java.io.IOException;
import java.io.InputStream;
import java.io.FileOutputStream;
import java.io.ByteArrayOutputStream;
import java.sql.SQLException;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import org.jdom.*;

import org.dspace.authorize.AuthorizeException;
import org.dspace.authorize.AuthorizeManager;
import org.dspace.content.DSpaceObject;
import org.dspace.content.Bitstream;
import org.dspace.content.Bundle;
import org.dspace.content.Item;
import org.dspace.content.Collection;
import org.dspace.content.DCValue;
import org.dspace.license.CreativeCommons;
import org.dspace.core.Constants;
import org.dspace.core.Context;
import org.dspace.core.Utils;
import org.dspace.core.PluginManager;
import org.dspace.handle.HandleManager;
import org.dspace.eperson.EPerson;
import org.dspace.content.packager.PackageException;
import org.dspace.content.packager.PackageValidationException;
import org.dspace.content.packager.PackageParameters;
import org.dspace.content.packager.PackageDisseminator;
import org.dspace.content.crosswalk.CrosswalkException;
import org.dspace.authorize.AuthorizeManager;

/**
 * This defines the behavior of DSpace "resources" in the WebDAV
 * interface; it maps DAV operations onto DSpace object.s
 */
class DAVItem extends DAVDSpaceObject
{
    /** log4j category */
    private static Logger log = Logger.getLogger(DAVItem.class);

    private Item item = null;

    private static final Element submitterProperty =
        new Element("submitter", DAV.NS_DSPACE);
    private static final Element getlastmodifiedProperty =
        new Element("getlastmodified", DAV.NS_DAV);
    private static final Element licenseProperty =
        new Element("license", DAV.NS_DSPACE);
    private static final Element cc_license_textProperty =
        new Element("cc_license_text", DAV.NS_DSPACE);
    private static final Element cc_license_rdfProperty =
        new Element("cc_license_rdf", DAV.NS_DSPACE);
    private static final Element cc_license_urlProperty =
        new Element("cc_license_url", DAV.NS_DSPACE);
    private static final Element owning_collectionProperty =
        new Element("owning_collection", DAV.NS_DSPACE);
    private static final Element withdrawnProperty =
        new Element("withdrawn", DAV.NS_DSPACE);

    private static List allProps = new Vector(commonProps);
    static
    {
        allProps.add(submitterProperty);
        allProps.add(getlastmodifiedProperty);
        allProps.add(licenseProperty);
        allProps.add(cc_license_textProperty);
        allProps.add(cc_license_rdfProperty);
        allProps.add(cc_license_urlProperty);
        allProps.add(owning_collectionProperty);
        allProps.add(handleProperty);
        allProps.add(withdrawnProperty);
    }

    protected List getAllProperties()
    {
        return allProps;
    }

    protected DAVItem(Context context,
                   HttpServletRequest request,
                   HttpServletResponse response,
                   String pathElt[],
                   Item item)
    {
        super(context, request, response, pathElt, item);
        this.type = TYPE_ITEM;
        this.item = item;
    }

    /**
     * Match Item URIs that identify the item by a database ID.
     * Handle URIs are matched by DAVDSpaceObject.
     */
    protected static DAVResource matchResourceURI(Context context,
                   HttpServletRequest request,
                   HttpServletResponse response,
                   String pathElt[])
        throws DAVStatusException, SQLException
    {
        int id = -1;
        String bsElt = null;

        try
        {
            // The "/item_db_" element in last or next-to-last element
            if (pathElt[pathElt.length-1].startsWith("item_db_"))
                id = Integer.parseInt(pathElt[pathElt.length-1].substring(8));
            else if (pathElt[pathElt.length-1].startsWith("bitstream_") &&
                     pathElt.length > 1 &&
                     pathElt[pathElt.length-2].startsWith("item_db_"))
            {
                id = Integer.parseInt(pathElt[pathElt.length-2].substring(8));
                bsElt = pathElt[pathElt.length-1];
            }
            if (id >= 0)
            {
                Item item = Item.find(context, id);
                if (item == null)
                    throw new DAVStatusException(HttpServletResponse.SC_NOT_FOUND, "Item with ID="+String.valueOf(id)+" not found.");
                if (bsElt != null)
                {
                    Bitstream bs = DAVBitstream.findBitstream(context, item, bsElt);
                    if (bs == null)
                        throw new DAVStatusException(HttpServletResponse.SC_NOT_FOUND, "Bitstream not found.");
                    return new DAVBitstream(context, request, response, pathElt, item, bs);
                }
                else
                    return new DAVItem(context, request, response, pathElt, item);
            }
            return null;
        }
        catch (NumberFormatException ne)
        {
            throw new DAVStatusException(HttpServletResponse.SC_BAD_REQUEST, "Error parsing number in request URI.");
        }
    }

    /**
     * Return pathname element in db-id format.
     */
    protected static String getPathElt(int dbid)
    {
        return "item_db_"+String.valueOf(dbid);
    }

    /**
     * Return pathname element to this Item.  Use db-id format
     * if no handle, otherwise the DSpace object format.
     */
    protected static String getPathElt(Item item)
    {
        String handle = item.getHandle();
        if (handle == null)
            return getPathElt(item.getID());
        else
            return DAVDSpaceObject.getPathElt(item);
    }


    /**
     * Return this resource's children.
     * Item's children are its bitstreams.
     */
    protected DAVResource[] children()
        throws SQLException
    {
        // Check for overall read permission on Item
        if (!AuthorizeManager.authorizeActionBoolean(context, item, Constants.READ))
            return new DAVResource[0];

        Vector result = new Vector();
        Bundle[] bundles = item.getBundles();
        for (int i = 0; i < bundles.length; i++)
        {
            // check read permission on this Bundle
            if (!AuthorizeManager.authorizeActionBoolean(context, bundles[i], Constants.READ))
                continue;

            Bitstream[] bitstreams = bundles[i].getBitstreams();
            for (int j = 0; j < bitstreams.length; j++)
            {
                String ext[] = bitstreams[j].getFormat().getExtensions();
                result.add(new DAVBitstream(context, request, response,
                             makeChildPath(DAVBitstream.getPathElt(
                                         bitstreams[j].getSequenceID(),
                                         ext.length < 1 ? null : ext[0])),
                             item, bitstreams[j]));
            }
        }
        return (DAVResource[]) result.toArray(new DAVResource[result.size()]);
    }

    protected Element propfindInternal(Element property)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        String value = null;

        /* FIXME:  This implements permission check that really belongs
         * in business logic.  Although communities and collections
         * don't check for read auth, Item may contain sensitive data
         * and should always check for READ permission.
         * Exception: allow "withdrawn" property to be checked regardless
         * of authorization since all permissions are removed when item
         * is withdrawn.
         */
        if (!elementsEqualIsh(property, withdrawnProperty))
            AuthorizeManager.authorizeAction(context, item, Constants.READ);

        if (elementsEqualIsh(property, withdrawnProperty))
            value = String.valueOf(item.isWithdrawn());

        else if (elementsEqualIsh(property, displaynameProperty))
        {
            // displayname - title or handle.
            DCValue titleDc[] = item.getDC("title", Item.ANY, Item.ANY);
            value = titleDc.length > 0 ? titleDc[0].value : item.getHandle();
        }
        else if (elementsEqualIsh(property, handleProperty))
            value = canonicalizeHandle(item.getHandle());

        else if (elementsEqualIsh(property, submitterProperty))
        {
            EPerson ep = item.getSubmitter();
            if (ep != null)
                value = hrefToEPerson(ep);
        }
        else if (elementsEqualIsh(property, owning_collectionProperty))
        {
            Collection owner = item.getOwningCollection();
            if (owner != null)
                value = canonicalizeHandle(owner.getHandle());
        }
        else if (elementsEqualIsh(property, getlastmodifiedProperty))
            value = DAV.HttpDateFormat.format(item.getLastModified());

        // dspace primary license
        else if (elementsEqualIsh(property, licenseProperty))
            value = getLicenseAsString();

        else if (elementsEqualIsh(property, cc_license_textProperty))
            value = CreativeCommons.getLicenseText(item);

        else if (elementsEqualIsh(property, cc_license_rdfProperty))
            value = CreativeCommons.getLicenseRDF(item);

        else if (elementsEqualIsh(property, cc_license_urlProperty))
            value = CreativeCommons.getLicenseURL(item);

        else
            return super.propfindInternal(property);

        // value was set up by "if" clause:
        if (value == null)
            throw new DAVStatusException(HttpServletResponse.SC_NOT_FOUND, "Not found.");
        Element p = new Element(property.getName(), property.getNamespace());
        p.setText(filterForXML(value));
        return p;
    }

    /**
     * Get the license for this Item as String.  License is the first
     * bitstream named "license.txt" in a LICENSE bundle, apparently?
     * <p>
     * FIXME: is this correct? there's no counterexample..
     * @return license string, or null if none found.
     */
    private String getLicenseAsString()
        throws SQLException, AuthorizeException, IOException
    {
        Bundle lb[] = item.getBundles(Constants.LICENSE_BUNDLE_NAME);
        for (int i = 0; i < lb.length; ++i)
        {
            Bitstream lbs = lb[i].getBitstreamByName("license.txt");
            if (lbs != null)
            {
                ByteArrayOutputStream baos = new ByteArrayOutputStream((int)lbs.getSize());
                Utils.copy(lbs.retrieve(), baos);
                return baos.toString();
            }
        }
        return null;
    }

    protected int proppatchInternal(int action, Element prop)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        // Don't need any authorization checks since the Item layer
        // checks authorization for write and delete.

        Namespace ns = prop.getNamespace();
        String propName = prop.getName();

        // "submitter" is the only Item-specific mutable property, rest are
        // live and unchangeable.
        if (ns != null && ns.equals(DAV.NS_DSPACE) &&
            propName.equals("submitter"))
        {
            if (action == DAV.PROPPATCH_REMOVE)
                throw new DAVStatusException(DAV.SC_CONFLICT,
                    "The submitter property cannot be removed.");
            String newName = prop.getText();
            EPerson ep = EPerson.findByEmail(context, newName);
            if (ep == null)
                throw new DAVStatusException(DAV.SC_CONFLICT,
                    "Cannot set submitter, no EPerson found for email address: "+newName);
            item.setSubmitter(ep);
            item.update();
            return HttpServletResponse.SC_OK;
        }
        throw new DAVStatusException(DAV.SC_CONFLICT, "The "+
            prop.getName()+" property cannot be changed.");
    }

    /**
     * GET implementation returns the contents of the Item as a package.
     * The query arg "package" must be specified.
     */
    protected void get()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        // Check for overall read permission on Item, because nothing else will
        AuthorizeManager.authorizeAction(context, item, Constants.READ);

        String packageType = request.getParameter("package");
        Bundle[] original = item.getBundles("ORIGINAL");
        int bsid;

        if (packageType == null)
            packageType = "default";
        PackageDisseminator dip = (PackageDisseminator)
          PluginManager.getNamedPlugin(PackageDisseminator.class, packageType);
        if (dip == null)
            throw new DAVStatusException(HttpServletResponse.SC_BAD_REQUEST, "Cannot find a disseminate plugin for package="+packageType);
        else
            try
            {
                PackageParameters pparams = PackageParameters.create(request);
                response.setContentType(dip.getMIMEType(pparams));
                dip.disseminate(context, item,
                                pparams,
                                response.getOutputStream());
            }
            catch (CrosswalkException pe)
            {
                throw new DAVStatusException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
                    "Failed in crosswalk of metadata: "+pe.toString());
            }
            catch (PackageException pe)
            {
                pe.log(log);
                throw new DAVStatusException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
                    pe.toString());
            }
    }

    protected void put()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED, "PUT is not implemented for Item.");
    }

    /**
     * COPY is implemented by an "add", which is the closest we can
     * get in DSpace semantics.
     */
    protected int copyInternal(DAVResource destination, int depth,
                                      boolean overwrite,
                                      boolean keepProperties)
        throws DAVStatusException, SQLException, AuthorizeException, IOException
    {
        return addItemToCollection(context, item, destination, overwrite);
    }

    /**
     * Do the work of "copy" method; this code is shared with e.g.
     * DAVInProgressSubmission.
     * @return HTTP status code.
     */
    protected static int addItemToCollection(Context context, Item item,
                                      DAVResource destination,
                                      boolean overwrite)
        throws DAVStatusException, SQLException, AuthorizeException, IOException
    {
        // sanity checks
        if (!(destination instanceof DAVCollection))
            throw new DAVStatusException(HttpServletResponse.SC_METHOD_NOT_ALLOWED,
                "COPY of Item is only allowed when destination is a DSpace Collection.");

        // access check
        AuthorizeManager.authorizeAction(context, item, Constants.READ);

        // make sure item doesn't belong to this collection
        Collection destColl = ((DAVCollection)destination).getCollection();

        log.debug("COPY from="+item.toString()+" ("+item.getHandle()+"), to="+
                    destColl.toString()+" ("+destColl.getHandle()+")");

        // check if it's already a member
        Collection refs[] = item.getCollections();
        for (int i = 0; i < refs.length; ++i)
            if (destColl.equals(refs[i]))
            {
                log.debug("COPY - item @ "+item.getHandle()+" is already a member of collection @ "+destColl.getHandle());
                if (overwrite)
                    return DAV.SC_NO_CONTENT;
                else
                    throw new DAVStatusException(DAV.SC_CONFLICT,
                      "This Item is already a member of collection handle="+destColl.getHandle());
            }
             
        destColl.addItem(item);
        return DAV.SC_NO_CONTENT;
    }
}
