/*
 * DAVSite.java
 *
 * Version: $Revision: 1.5 $
 *
 * Date: $Date: 2005/04/20 14:22:36 $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.app.dav;

import java.io.IOException;
import java.sql.SQLException;
import java.net.URLDecoder;
import java.util.List;
import java.util.Vector;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import org.jdom.*;

import org.dspace.authorize.AuthorizeException;
import org.dspace.authorize.AuthorizeManager;
import org.dspace.content.Community;
import org.dspace.core.Constants;
import org.dspace.core.Context;
import org.dspace.core.ConfigurationManager;
import org.dspace.handle.HandleManager;

/**
 * Model the DSpace Site as a resource.
 * The URI is the top level "collection".  Only an
 * Administrator can modify properties.
 */
class DAVSite extends DAVResource
{
    /** log4j category */
    private static Logger log = Logger.getLogger(DAVSite.class);

    private static final Element news_topProperty =
        new Element("news_top", DAV.NS_DSPACE);
    private static final Element news_sideProperty =
        new Element("news_side", DAV.NS_DSPACE);
    private static final Element default_licenseProperty =
        new Element("default_license", DAV.NS_DSPACE);

    private static List allProps = new Vector(commonProps);
    static
    {
        allProps.add(news_topProperty);
        allProps.add(news_sideProperty);
        allProps.add(default_licenseProperty);
    }

    protected Element typeValue()
    {
        return new Element("site", DAV.NS_DSPACE);
    }

    protected DAVSite(Context context,
                   HttpServletRequest request,
                   HttpServletResponse response,
                   String pathElt[])
    {
        super(context, request, response, pathElt);
        this.type = TYPE_SITE;
    }

    protected static DAVResource matchResourceURI(Context context,
                   HttpServletRequest request,
                   HttpServletResponse response,
                   String pathElt[])
        throws DAVStatusException, SQLException
    {
        if (pathElt.length == 0 || pathElt[0].length() == 0)
            return new DAVSite(context, request, response, new String[0]);
        else
            return null;
    }

    protected List getAllProperties()
    {
        return allProps;
    }

    protected DAVResource[] children()
        throws SQLException
    {
        Community top[] = Community.findAllTop(context);
        DAVResource result[] = new DAVResource[top.length];
         
        for (int i = 0; i < top.length; ++i)
            result[i] = new DAVCommunity(context, request, response,
                                        makeChildPath(top[i]), top[i]);
        return result;
    }

    protected Element propfindInternal(Element property)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        String value = null;

        // displayname - title or handle.
        if (elementsEqualIsh(property, displaynameProperty))
            value = ConfigurationManager.getProperty("dspace.name");

        else if (elementsEqualIsh(property, news_topProperty))
            value = ConfigurationManager.readNewsFile(Constants.NEWS_TOP);

        else if (elementsEqualIsh(property, news_sideProperty))
            value = ConfigurationManager.readNewsFile(Constants.NEWS_SIDE);

        else if (elementsEqualIsh(property, default_licenseProperty))
            value = ConfigurationManager.getDefaultSubmissionLicense();

        else
            return commonPropfindInternal(property, true);

        // value was set up by "if" clause:
        if (value == null)
            throw new DAVStatusException(HttpServletResponse.SC_NOT_FOUND, "Not found.");
        Element p = new Element(property.getName(), property.getNamespace());
        p.setText(filterForXML(value));
        return p;
    }

    protected int proppatchInternal(int action, Element prop)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        String newValue = (action == DAV.PROPPATCH_REMOVE) ? null :
                            prop.getText();
        if (elementsEqualIsh(prop, news_topProperty))
        {
            if (!AuthorizeManager.isAdmin(context))
                throw new DAVStatusException(HttpServletResponse.SC_FORBIDDEN,
                    "Not authorized to modify this property.");
            ConfigurationManager.writeNewsFile(Constants.NEWS_TOP, newValue);
        }
        else if (elementsEqualIsh(prop, news_sideProperty))
        {
            if (!AuthorizeManager.isAdmin(context))
                throw new DAVStatusException(HttpServletResponse.SC_FORBIDDEN,
                    "Not authorized to modify this property.");
            ConfigurationManager.writeNewsFile(Constants.NEWS_SIDE, newValue);
        }
        else if (elementsEqualIsh(prop, displaynameProperty))
            throw new DAVStatusException(DAV.SC_CONFLICT,
                "The site name can only be changed through the DSpace Configuration, \"dspace.name\" property.");
        else
            throw new DAVStatusException(DAV.SC_CONFLICT, "The "+
                prop.getName()+" property cannot be changed.");
        return HttpServletResponse.SC_OK;
    }

    protected void get()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED, "GET is not implemented for Site.");
    }


    protected void put()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED, "PUT is not implemented for Site.");
    }


    protected int copyInternal(DAVResource destination, int depth,
                                      boolean overwrite,
                                      boolean keepProperties)
        throws DAVStatusException, SQLException, AuthorizeException, IOException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED,
            "COPY method not implemented.");
    }
}
