/*
 * DAVWorkflow.java
 *
 * Version: $Revision: 1.5 $
 *
 * Date: $Date: 2005/04/20 14:22:36 $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.app.dav;

import java.io.IOException;
import java.sql.SQLException;
import java.util.List;
import java.util.Vector;
import java.util.ListIterator;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import org.jdom.Element;

import org.dspace.authorize.AuthorizeException;
import org.dspace.core.Constants;
import org.dspace.core.Context;
import org.dspace.eperson.EPerson;
import org.dspace.content.WorkspaceItem;
import org.dspace.content.InProgressSubmission;
import org.dspace.workflow.WorkflowItem;
import org.dspace.workflow.WorkflowManager;

/**
 * The "workflow_*" resources are collections (in the DAV sense) of all
 * the current user's <code>WorkflowItem</code>s.  Each one is nothing more
 * than a read-only collection to list the items.
 * <p>
 * The <code>workflow_own</code> resource lists items <em>owned</em> by
 * the current user, while the <code>workflow_pool</code> lists items
 * in the workflow pool which may be acquired by the current user.
 * <p>
 * Its children are all the relevant <code>WorkflowItem</code>s.
 * These resources cannot be altered.
 * <p>
 * @author Larry Stone
 * @see DAVWorkflowItem
 */
class DAVWorkflow extends DAVResource
{
    /** log4j category */
    private static Logger log = Logger.getLogger(DAVWorkflow.class);

    private static List allProps = new Vector(commonProps);

    protected Element typeValue()
    {
        return new Element("workflow", DAV.NS_DSPACE);
    }

    protected DAVWorkflow(Context context,
                   HttpServletRequest request,
                   HttpServletResponse response,
                   String pathElt[])
    {
        super(context, request, response, pathElt);
        type = TYPE_OTHER;
    }

    /**
     * @return URI path to this object.
     */
    protected static String getPath(WorkflowItem wfi)
    {
        return "workflow_pool/"+DAVWorkflowItem.getPathElt(wfi);
    }

    /**
     * Match the URIs this subclass understands and return the
     * corresponding resource.
     */
    protected static DAVResource matchResourceURI(Context context,
                   HttpServletRequest request,
                   HttpServletResponse response,
                   String pathElt[])
        throws DAVStatusException, SQLException
    {
        // The "/workflow" URI:
        if (pathElt.length > 0 && pathElt[0].startsWith("workflow_"))
        {
            if (pathElt.length > 1)
                return DAVWorkflowItem.matchResourceURI(context, request,
                            response,pathElt);
            else if (pathElt[0].equals("workflow_own") ||
                     pathElt[0].equals("workflow_pool"))
                return new DAVWorkflow(context, request, response, pathElt);
            else
                throw new DAVStatusException(HttpServletResponse.SC_NOT_FOUND, "Unrecognized URI path element: "+pathElt[0]);
        }
        return null;
    }

    protected List getAllProperties()
    {
        return allProps;
    }

    protected DAVResource[] children()
        throws SQLException
    {


        EPerson ep = context.getCurrentUser();
        if (ep != null)
        {
            List wi = null;
            if (pathElt[0].equals("workflow_own"))
                wi = WorkflowManager.getOwnedTasks(context, ep);
            else if (pathElt[0].equals("workflow_pool"))
                wi = WorkflowManager.getPooledTasks(context, ep);
            if (wi != null)
            {
                log.debug("children(): Got "+String.valueOf(wi.size())+" Workflow Items.");
                DAVResource result[] = new DAVResource[wi.size()];
                ListIterator wii = wi.listIterator();
                int i = 0;
                while (wii.hasNext())
                {
                    WorkflowItem wfi = (WorkflowItem)wii.next();
                    result[i++] =
                      new DAVWorkflowItem(context, request, response,
                        makeChildPath(DAVWorkflowItem.getPathElt(wfi.getID())),
                        wfi);
                }
                return result;
            }
        }
        return new DAVResource[0];
    }

    protected Element propfindInternal(Element property)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        String value = null;

        // displayname - title or handle.
        if (elementsEqualIsh(property, displaynameProperty))
            value = pathElt[0];

        else
            return commonPropfindInternal(property, true);

        // value was set up by "if" clause:
        if (value == null)
            throw new DAVStatusException(HttpServletResponse.SC_NOT_FOUND, "Not found.");
        Element p = new Element(property.getName(), property.getNamespace());
        p.setText(filterForXML(value));
        return p;
    }

    protected int proppatchInternal(int action, Element prop)
        throws SQLException, AuthorizeException, IOException, DAVStatusException
    {
        throw new DAVStatusException(DAV.SC_CONFLICT, "The "+
            prop.getName()+" property cannot be changed.");
    }

    protected void get()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED,
                    "GET method not implemented for workflow.");
    }

    protected void put()
        throws SQLException, AuthorizeException, ServletException, IOException,
               DAVStatusException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED,
                    "PUT method not implemented for workflow.");
    }

    protected int copyInternal(DAVResource destination, int depth,
                                      boolean overwrite,
                                      boolean keepProperties)
        throws DAVStatusException, SQLException, AuthorizeException, IOException
    {
        throw new DAVStatusException(HttpServletResponse.SC_NOT_IMPLEMENTED,
            "COPY method not implemented for workflow.");
    }
}
