/*
 * LNISoapServlet.java
 *
 * Version: $Revision: 1.5 $
 *
 * Date: $Date: 2005/04/20 14:22:36 $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.app.dav;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.FileOutputStream;
import java.sql.SQLException;
import java.util.List;
import java.util.Iterator;

import javax.servlet.GenericServlet;
import javax.servlet.ServletException;
import javax.servlet.ServletConfig;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServlet;

import org.apache.log4j.Logger;

import org.jdom.*;
import org.jdom.output.XMLOutputter;
import org.jdom.output.Format;

import org.apache.axis.transport.http.AxisServlet;
import org.apache.axis.MessageContext;

import org.dspace.authorize.AuthorizeException;
import org.dspace.core.Context;
import org.dspace.core.Utils;
import org.dspace.eperson.EPerson;
import org.dspace.eperson.AuthenticationManager;
import org.dspace.eperson.AuthenticationMethod;

/**
 * Servlet implementing SOAP services of DSpace Lightweight Network Interface
 * <P>
 * This is implemented as a subclass of the AxisServlet that
 * processes SOAP requests, so it can pick out the requests it handles
 * and pass on the rest to the Axis Engine.
 * <p>
 * Note that it also handles WebDAV GET and PUT requests, so the SOAP client
 * can use the same URL as a SOAP endpoint and WebDAV resource root.
 *
 * @author Larry Stone
 * @version $Revision: 1.5 $
 */
public class LNISoapServlet extends AxisServlet

{
    /** log4j category */
    private static Logger log = Logger.getLogger(LNISoapServlet.class);

    private static XMLOutputter outputPretty = new XMLOutputter(Format.getPrettyFormat());

    // state of this transaction
    private HttpServletRequest request = null;
    private HttpServletResponse response = null;

    // last servlet instance when put into service, set by init()
    private static GenericServlet servletInstance = null;

    public void init(ServletConfig sc)
        throws ServletException
    {
        super.init(sc);
        servletInstance = this;
    }

    public static GenericServlet getServletInstance()
    {
        return servletInstance;
    }

    /**
     * Pass a GET request directly to the WebDAV implementation.
     * It handles authentication.
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException
    {
        DAVServlet.serviceInternal("GET", request, response);
    }

    /**
     * Pass a PUT request directly to the WebDAV implementation.
     * It handles authentication.
     */
    public void doPut(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException
    {
        DAVServlet.serviceInternal("PUT", request, response);
    }

    /**
     * Authenticate and return the filled-in DSpace context
     * This is the prologue to all calls.
     */
    private Context prologue()
        throws SQLException, IOException
    {
        MessageContext mc = MessageContext.getCurrentContext();
        String username = null, password = null;

        if (mc.getUsername() != null)
            username = DAVServlet.URLDecode(mc.getUsername());
        if (mc.getPassword() != null)
            password = DAVServlet.URLDecode(mc.getPassword());

        /**** XXX TEMPORARY
         **** Instrumentation to explore the guts of Axis at runtime,
         **** leave this commented-out.
        java.util.Iterator pi = mc.getPropertyNames();
        while (pi.hasNext())
            log.debug("SOAP: request has property named \""+((String)pi.next())+"\"");
        log.debug("SOAP: getSOAPActionURI = \""+mc.getSOAPActionURI()+"\"");
        log.debug("SOAP: property(servletEndpointContext) =  = \""+mc.getProperty("servletEndpointContext").toString()+"\"");
        log.debug("SOAP: property(realpath) =  = \""+mc.getProperty("realpath").toString()+"\"");
        log.debug("SOAP: property(transport.http.servletLocation) =  = \""+mc.getProperty("transport.http.servletLocation").toString()+"\"");
         **** end TEMPORARY INSTRUMENTATION
         ****/

        request = (HttpServletRequest)
                mc.getProperty("transport.http.servletRequest");
        response = (HttpServletResponse)
                mc.getProperty("transport.http.servletResponse");
        Context context = new Context();

        // try cookie shortcut
        if (DAVServlet.getAuthFromCookie(context, request))
        {
            DAVServlet.putAuthCookie(context, request, response, false);
            log.debug("SOAP service "+this.getClass().getName()+
                 " authenticated with cookie.");
            return context;
        }

        int status = AuthenticationManager.authenticate(context, username,
                        password, null, request);
        if (status == AuthenticationMethod.SUCCESS)
        {
            EPerson cu = context.getCurrentUser();
            log.debug("SOAP service "+this.getClass().getName()+
                 " authenticated as "+cu.getEmail()+
                 " ("+cu.getFirstName()+" "+cu.getLastName()+")");
            DAVServlet.putAuthCookie(context, request, response, true);
            return context;
        }
        else if (status == AuthenticationMethod.BAD_CREDENTIALS)
            throw new LNIRemoteException("Authentication failed: Bad Credentials.");
        else if (status == AuthenticationMethod.CERT_REQUIRED)
            throw new LNIRemoteException("Authentication failed: This user may only login with X.509 certificate.");
        else if (status == AuthenticationMethod.NO_SUCH_USER)
            throw new LNIRemoteException("Authentication failed: No such user.");
        else    /** AuthenticationMethod.BAD_ARGS  and etc **/
            throw new LNIRemoteException("Authentication failed: Cannot authenticate.");
    }

    public String propfind(String uri, String doc, int depth, String types)
        throws LNIRemoteException
    {
        // break up path into elements.
        if (uri.startsWith("/"))
            uri = uri.substring(1);
        String pathElt[] = uri.split("/");

        try
        {
            Context context = prologue();

            // return properties only for resources of these types, comma-sep list
            String aTypes[] = (types == null) ? null : types.split(",");
            int typeMask = DAVResource.typesToMask(aTypes);
             
            DAVResource resource = DAVResource.findResource(context, null, null, pathElt);
            if (resource == null)
                throw new LNIRemoteException("Resource not found.");
            else
            {
                Document outdoc = resource.propfindDriver(depth,
                                        new ByteArrayInputStream(doc.getBytes()),
                                        typeMask);
                if (outdoc != null)
                {
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    outputPretty.output(outdoc, baos);
                    return baos.toString();
                }
                else
                {
                    // this should never happen, it should throw an error
                    // before returning null
                    throw new LNIRemoteException("propfind failed, no document returned.");
                }
            }
        }
        catch (IOException ie)
        {
            throw new LNIRemoteException("Exception executing PROPFIND", ie);
        }
        catch (SQLException e)
        {
            throw new LNIRemoteException("Failure accessing database", e);
        }
        catch (DAVStatusException e)
        {
            throw new LNIRemoteException("PROPFIND request failed: "+e.getStatusLine());
        }
        catch (AuthorizeException e)
        {
            throw new LNIRemoteException("You are not authorized for the requested operation.", e);
        }
    }

    public String proppatch(String uri, String doc)
        throws LNIRemoteException
    {
        // break up path into elements.
        if (uri.startsWith("/"))
            uri = uri.substring(1);
        String pathElt[] = uri.split("/");

        try
        {
            Context context = prologue();
            DAVResource resource = DAVResource.findResource(context, null, null, pathElt);
            if (resource == null)
                throw new LNIRemoteException("Resource not found.");
            else
            {
                Document outdoc = resource.proppatchDriver(
                                    new ByteArrayInputStream(doc.getBytes()));
                if (outdoc != null)
                {
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    outputPretty.output(outdoc, baos);
                    return baos.toString();
                }
                else
                {
                    // this should never happen, it should throw an error
                    // before returning null
                    throw new LNIRemoteException("proppatch failed, no document returned.");
                }
            }
        }
        catch (IOException ie)
        {
            throw new LNIRemoteException("Exception executing PROPPATCH", ie);
        }
        catch (SQLException e)
        {
            throw new LNIRemoteException("Failure accessing database", e);
        }
        catch (DAVStatusException e)
        {
            throw new LNIRemoteException("PROPPATCH request failed: "+e.getStatusLine());
        }
        catch (AuthorizeException e)
        {
            throw new LNIRemoteException("You are not authorized for the requested operation.", e);
        }
    }

    private static String lookupPathElt[] = {"lookup", "handle"};

    /**
     * Return "absolute" DAV URI for the given handle (and optional
     * bitstream persistent identifier).
     * Always returns a valid URI; if resource is not found it
     * throws an exception.
     */
    public String lookup(String handle, String bitstreamPid)
        throws LNIRemoteException
    {
        try
        {
            Context context = prologue();

            // trim leading scheme if any:
            if (handle.startsWith("hdl:"))
                handle = handle.substring(4);

            DAVLookup resource = new DAVLookup(context, request, response, lookupPathElt);
            String result = resource.makeURI(handle, bitstreamPid);
            if (result == null)
                throw new LNIRemoteException("Resource not found.");
            return result;
        }
        catch (IOException ie)
        {
            throw new LNIRemoteException("Exception executing LOOKUP", ie);
        }
        catch (SQLException e)
        {
            throw new LNIRemoteException("Failure accessing database", e);
        }
    }


    public int copy(String source, String destination, int depth,
                    boolean overwrite, boolean keepProperties)
        throws LNIRemoteException
    {
        // break up path into elements.
        if (source.startsWith("/"))
            source = source.substring(1);
        String pathElt[] = source.split("/");

        try
        {
            Context context = prologue();
            DAVResource resource = DAVResource.findResource(context, null, null, pathElt);
            if (resource == null)
                throw new LNIRemoteException("Resource not found.");

            return resource.copyDriver(destination, depth, overwrite, keepProperties);
        }
        catch (IOException ie)
        {
            throw new LNIRemoteException("IOException while executing COPY", ie);
        }
        catch (SQLException e)
        {
            throw new LNIRemoteException("Failure accessing database", e);
        }
        catch (DAVStatusException e)
        {
            throw new LNIRemoteException("COPY request failed: "+e.getStatusLine());
        }
        catch (AuthorizeException e)
        {
            throw new LNIRemoteException("You are not authorized for the requested operation.", e);
        }
    }
}
