/*
 * RedirectFilter.java
 *
 * Version: $Revision: 1.5 $
 *
 * Date: $Date: 2005/04/20 14:22:37 $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.app.filter;

import java.io.IOException;
import java.io.InputStream;

import java.net.URL;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.RequestDispatcher;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * Simple Servlet Container filter redirect requests to another
 * Webapp and server on the same server.
 * <p>
 * It tries two different techniques to do the redirect:
 * <br> 1. Forward the request internally to a different servlet.  This
 * is more efficient but may not work on all servlet containers.
 *
 * <br> 2. Issue a HTTP "redirect" response directing the client to
 * try a different URL.  This always works but it is less efficient.
 *
 * <p>
 * This filter must be configured in the webapp's deployment descriptor.
 * It recognizes the following init parameters:
 * <pre>
 * A. "targetURI", "targetServlet" -- control the Internal Forwarding method.
 *
 *       *** NOTE ***
 *       *** NOTE ***
 *       *** NOTE ***
 *    To use Internal Forwarding under Tomcat 5.0 (and later?), you MUST
 *    enable cross-context access, _and_ _be_ _sure_ you are not compromising
 *    the security of your web server in doing so.  To do this, add the
 *    following element within the <Engine> element in your conf/servlet.xml:
 *
 *      <DefaultContext crossContext="true" />
 *
 *    See the Apache Jakarta Tomcat Server Configuration Reference for
 *    more details.
 *
 *    Other servlet containers, such as Jetty, may have other rules about
 *    enabling and disabling the use of cross-webapp getContext().
 *    (Jetty 5.x seems to allow cross-webapp context access by default.)
 *       *** END of NOTE ***
 *
 *    The two parameters "targetURI" and "targetServlet" must be
 *    specified together:
 *
 *      1. "targetURI" is the the _complete_ URI (i.e.  the full URL
 *         after the host component) of the path you want to redirect _to_.
 *      2. "targetServlet" is the servlet component of that URI, i.e.
 *         the URI _after_ the servlet context path.
 *
 *      Example: If you are redirecting _to_ the URL
 *       http://uni.edu/dspace/oai/request, and the dspace webapp is mounted
 *       at the URI "/dspace", then:
 *        <init-param>
 *          <param-name>targetURI</param-name>
 *          <param-value>/dspace/oai/request</param-value>
 *        </init-param>
 *        <init-param>
 *          <param-name>targetServlet</param-name>
 *          <param-value>/oai/request</param-value>
 *        </init-param>
 *
 *      NOTE: targetServlet must include the complete servlet path as it
 *      would be part of a URL directed to the servlet.
 *
 *  B. "match", "replace" -- control the HTTP Redirect method.
 *
 *      The two parameters "match" and "replace" work together:
 *      Set "match" to the exact string (NOT regexp) to match in request URL
 *      Set "replace" takes the place of the "match" string in the redirect URL -->
 *
 *      This example rewrites "/dspace-oai/request" in a URI to be
 *      "/dspace/oai/request", which will send
 *
 *          <init-param>
 *            <param-name>match</param-name>
 *            <param-value>/dspace-oai/request</param-value>
 *          </init-param>
 *          <init-param>
 *            <param-name>replace</param-name>
 *            <param-value>/dspace/oai/request</param-value>
 *          </init-param>
 *  </pre>
 *
 * @author Larry Stone
 */
public class RedirectFilter implements Filter
{
    // The servlet context of the webapp we're running under.
    private ServletContext context;

    // match this string
    private String match = null;

    // replace with this
    private String replace = null;

    // target URI to which to redirect, e.g. /dspace/oai (for webapp at /dspace)
    private String targetURI = null;

    // servlet portion of target URI path, e.g. /oai
    private String targetServlet = null;

    /**
     * Initialize this filter.  Load config parameters since it only
     * has to be done once.
     *
     * @param filterConfig The filter config
     * @throws ServletException If an error occurs
     */
    public void init(FilterConfig filterConfig)
        throws ServletException
    {
        context = filterConfig.getServletContext();

        match = filterConfig.getInitParameter("match");
        if (match == null)
            throw new ServletException("init-param match is required.");

        replace = filterConfig.getInitParameter("replace");
        if (replace == null)
            throw new ServletException("init-param replace is required.");

        // this is allowed to be null.
        targetURI = filterConfig.getInitParameter("targetURI");
        targetServlet = filterConfig.getInitParameter("targetServlet");
    }


    /**
     * Forward an HTTP request to the LNI WebDAV servlet.  Handle
     * GET and PUT methods by forwarding them, rather than attempting
     * to duplicate
     * @return true if successfully forwarded, false otherwise
     */
    private boolean forwardToTarget(ServletRequest request, ServletResponse response)
        throws ServletException, IOException
    {
        if (targetURI == null || targetServlet == null)
            return false;

        ServletContext tc = context.getContext(targetURI);
        if (tc == null)
        {
            context.log(
                getClass().getName()+": DEBUG:"+
              "Aborting, cannot get ServletContext for URI="+targetURI);
            return false;
        }

        RequestDispatcher rd = tc.getRequestDispatcher(targetServlet);
        if (rd == null)
        {
            context.log(
                getClass().getName()+": DEBUG:"+
                      "Aborting, cannot get RequestDispatcher for URI="+targetURI+", servlet="+targetServlet);
            return false;
        }
        rd.forward(request,response);
        return true;
    }

    /**
     * Filter the incoming request by redirecting it according to config.
     * First try to forward it internally if the targetURI and targetServlet
     * are configured.  If that is impossible, continue to attempt
     * to redirect by matching and replacing the "match" and "replace"
     * strings, and if that succeeds, send an HTTP redirect response.
     *
     * If none of the redirect parameters succeeds in making a match, pass
     * the request along, in case we are only filtering some requests.
     *
     * @param request The request
     * @param response The response
     * @param chain The filter chain
     * @throws IOException If an error occurs
     * @throws ServletException If an error occurs
     */
    public void doFilter(
        ServletRequest request, ServletResponse response, FilterChain chain)
        throws IOException, ServletException
    {
        if (response.isCommitted())
        {
            context.log(
                getClass().getName()
                + ": doFilter(): not inspecting committed response.");
            chain.doFilter(request, response);
        }
        else if (!(request instanceof HttpServletRequest))
        {
            context.log(
                getClass().getName()
                + ": doFilter(): not inspecting non-Http request.");
            chain.doFilter(request, response);
        }

        // First try forwarding through servlet context, if server allows.
        else if (forwardToTarget(request, response))
        {
            context.log(
                getClass().getName()+": DEBUG:"+
                      "forwarded through Context to targetURI="+targetURI);
        }

        // try to rewrite the URI and redirect
        else
        {
            HttpServletRequest hreq = (HttpServletRequest) request;
            HttpServletResponse hres = (HttpServletResponse) response;

            String reqUrl = hreq.getRequestURL().toString();
            String query = hreq.getQueryString();
            if (query != null)
                reqUrl = reqUrl + "?" + query;
            int s = reqUrl.indexOf(match);
            if (s >= 0)
            {
                String newUrl = reqUrl.substring(0, s) +
                                replace + reqUrl.substring(s+match.length());

                String encoded = hres.encodeRedirectURL(newUrl);

                context.log(
                    getClass().getName()
                    + ": doFilter(): redirecting request for " + reqUrl
                    + " to " + newUrl);

                hres.sendRedirect(encoded);
            } else {
                chain.doFilter(request, response);
            }
        }
    }

    public void destroy()
    {
        context = null;
    }
}
